@@ -11,41 +11,62 @@
 #ifndef P_tmpdir
 #define P_tmpdir "/tmp"
 #endif
+
+/* NOTE: The code below is #include'd both from a plain C program (boot.c)
+ * and our custom Perl interpreter (main.c). In the latter case,
+ * lstat() or stat() may be #define'd as calls into PerlIO and
+ * expect &PL_statbuf as second parameter, rather than a pointer
+ * to a struct stat. Try to distinguish these cases by checking
+ * whether PL_statbuf is defined. */
 static int isWritableDir(const char* val)
 {
-    /* NOTE: This code is #include'd both from a plain C program (static.c)
-     * and our custom Perl interpreter (main.c). In the latter case,
-     * lstat() or stat() may be #define'd as calls into PerlIO and
-     * expect &PL_statbuf as second parameter, rather than a pointer
-     * to a struct stat. Try to distinguish these cases by checking
-     * whether PL_statbuf is defined. */
 #ifndef PL_statbuf
     struct stat PL_statbuf;
 #endif
 
     return par_lstat(val, &PL_statbuf) == 0 && 
-             ( S_ISDIR(PL_statbuf.st_mode) ||
-               S_ISLNK(PL_statbuf.st_mode) ) &&
-             access(val, W_OK) == 0;
+           ( S_ISDIR(PL_statbuf.st_mode) || S_ISLNK(PL_statbuf.st_mode) ) &&
+           access(val, W_OK) == 0;
 }
 
+#ifndef WIN32
+/* check that:
+ * - val is a directory (and not a symlink)
+ * - val is owned by the user
+ * - val has mode 0700
+ */
+static int isSafeDir(const char* val)
+{
+#ifndef PL_statbuf
+    struct stat PL_statbuf;
+#endif
+
+    return par_lstat(val, &PL_statbuf) == 0 && 
+           S_ISDIR(PL_statbuf.st_mode) &&
+           PL_statbuf.st_uid == getuid() &&
+           (PL_statbuf.st_mode & 0777) == 0700;
+}
+#endif
+
 void par_setup_libpath( const char * stmpdir )
 {
     const char *key = NULL , *val = NULL;
     int i;
     char *ld_path_env = NULL;
 
-    /* NOTE: array is terminated by an empty string */
+    /* NOTE: array is NULL terminated */
     const char *ld_path_keys[] = {       
         "LD_LIBRARY_PATH", "LIBPATH", "LIBRARY_PATH",
-        "PATH", "DYLD_LIBRARY_PATH", "SHLIB_PATH", ""
+        "PATH", "DYLD_LIBRARY_PATH", "SHLIB_PATH", NULL
     };
 
-    for ( i = 0 ; strlen(key = ld_path_keys[i]) > 0 ; i++ ) {
-        if ( ((val = par_getenv(key)) == NULL) || (strlen(val) == 0) ) {
+    /* parentheses around the "for" condition are there to shut up compiler warnings */
+    for ( i = 0 ; (key = ld_path_keys[i]); i++ ) {
+        if ( (val = par_getenv(key)) == NULL || strlen(val) == 0 ) {
             par_setenv(key, stmpdir);
         }
-        else if(!strstr(val, stmpdir)) {
+        else if ( !strstr(val, stmpdir) ) {
+            /* prepend stmpdir to (value of) environment variable */
             ld_path_env = malloc( 
 		strlen(stmpdir) + strlen(path_sep) + strlen(val) + 1);
             sprintf(
@@ -58,26 +79,25 @@ void par_setup_libpath( const char * stmpdir )
 
 char *par_mktmpdir ( char **argv ) {
     int i;
-    char *c;
     const char *tmpdir = NULL;
     const char *key = NULL , *val = NULL;
 
-    /* NOTE: all arrays below are terminated by an empty string */
+    /* NOTE: all arrays below are NULL terminated */
     const char *temp_dirs[] = { 
         P_tmpdir, 
 #ifdef WIN32
         "C:\\TEMP", 
 #endif
-        ".", "" };
+        ".", NULL };
     const char *temp_keys[] = { "PAR_TMPDIR", "TMPDIR", "TEMPDIR", 
-                                 "TEMP", "TMP", "" };
-    const char *user_keys[] = { "USER", "USERNAME", "" };
+                                 "TEMP", "TMP", NULL };
+    const char *user_keys[] = { "USER", "USERNAME", NULL };
 
     const char *subdirbuf_prefix = "par-";
     const char *subdirbuf_suffix = "";
 
     char *progname = NULL, *username = NULL;
-    char *stmpdir = NULL, *stmpdir2 = NULL;
+    char *stmpdir = NULL, *top_tmpdir = NULL;
     int f, j, k, stmp_len = 0;
     char sha1[41];
     SHA_INFO sha_info;
@@ -94,45 +114,41 @@ char *par_mktmpdir ( char **argv ) {
         DWORD buflen = MAXPATHLEN;
         username = malloc(MAXPATHLEN);
         GetUserName((LPTSTR)username, &buflen);
+        // FIXME this is uncondifionally overwritten below - WTF?
     }
 #endif
 
     /* Determine username */
     username = get_username_from_getpwuid();
-    if ( username == NULL ) { /* fall back to env vars */
-        for (
-                i = 0 ;
-                username == NULL && strlen(key = user_keys[i]) > 0 ;
-                i++
-            )
-        {
-            if ( (val = par_getenv(key)) ) username = strdup(val);
+    if ( !username ) { /* fall back to env vars */
+        for ( i = 0 ; username == NULL && (key = user_keys[i]); i++) {
+            if ( (val = par_getenv(key)) && strlen(val) ) 
+                username = strdup(val);
         }
     }
-
-    if ( username == NULL ) {
+    if ( username == NULL )
         username = "SYSTEM";
-    }
-    else {
-        /* replace all non-alphanumeric letters with '_' */
-        for ( c = username ; *c != '\0' ; c++ ) {
-            if ( !isalnum(*c) ) {
-                *c = '_';
-            }
-        }
+   
+    /* sanitize username: encode all bytes as 2 hex digits */
+    {
+        char *hexname = malloc(2 * strlen(username) + 1);
+        char *u, *h;
+        for ( u = username, h = hexname ; *u != '\0' ; u++, h += 2)
+            sprintf(h, "%02x", *(unsigned char*)u);
+        username = hexname;
     }
 
     /* Try temp environment variables */
-    for ( i = 0 ; tmpdir == NULL && strlen(key = temp_keys[i]) > 0 ; i++ ) {
-        if ( (val = par_getenv(key)) &&
-                isWritableDir(val) ) {
+    for ( i = 0 ; tmpdir == NULL && (key = temp_keys[i]); i++ ) {
+        if ( (val = par_getenv(key)) && strlen(val) && isWritableDir(val) ) {
             tmpdir = strdup(val);
+            break;
         }
     }
 
 #ifdef WIN32
     /* Try the windows temp directory */
-    if ( tmpdir == NULL && (val = par_getenv("WinDir")) ) {
+    if ( tmpdir == NULL && (val = par_getenv("WinDir")) && strlen(val) ) {
         char* buf = malloc(strlen(val) + 5 + 1);
         sprintf(buf, "%s\\temp", val);
         if (isWritableDir(buf)) {
@@ -144,7 +160,7 @@ char *par_mktmpdir ( char **argv ) {
 #endif
 
     /* Try default locations */
-    for ( i = 0 ; tmpdir == NULL && strlen(val = temp_dirs[i]) > 0 ; i++ ) {
+    for ( i = 0 ; tmpdir == NULL && (val = temp_dirs[i]) && strlen(val) ; i++ ) {
         if ( isWritableDir(val) ) {
             tmpdir = strdup(val);
         }
@@ -157,14 +173,31 @@ char *par_mktmpdir ( char **argv ) {
         strlen(username) +
         strlen(subdirbuf_suffix) + 1024;
 
-    /* stmpdir is what we are going to return 
-       stmpdir2 is the top $TEMP/par-$USER, needed to build stmpdir.  We
-       need 2 buffers because snprintf() can't write to a buffer it's
-       reading from. */
+    /* stmpdir is what we are going to return; 
+       top_tmpdir is the top $TEMP/par-$USER, needed to build stmpdir.  
+       NOTE: We need 2 buffers because snprintf() can't write to a buffer
+       it is also reading from. */
+    top_tmpdir = malloc( stmp_len );
+    sprintf(top_tmpdir, "%s%s%s%s", tmpdir, dir_sep, subdirbuf_prefix, username);
+#ifdef WIN32
+    _mkdir(top_tmpdir);         /* FIXME bail if error (other than EEXIST) */
+#else
+    {
+        if (mkdir(top_tmpdir, 0700) == -1 && errno != EEXIST) {
+            fprintf(stderr, "%s: creation of private subdirectory %s failed (errno=%i)\n", 
+                    argv[0], top_tmpdir, errno);
+            return NULL;
+        }
+
+        if (!isSafeDir(top_tmpdir)) {
+            fprintf(stderr, "%s: private subdirectory %s is unsafe (please remove it and retry your operation)\n",
+                    argv[0], top_tmpdir);
+            return NULL;
+        }
+    }
+#endif
+
     stmpdir = malloc( stmp_len );
-    stmpdir2 = malloc( stmp_len );
-    sprintf(stmpdir2, "%s%s%s%s", tmpdir, dir_sep, subdirbuf_prefix, username);
-    my_mkdir(stmpdir2, 0755);
 
     /* Doesn't really work - XXX */
     val = par_getenv( "PATH" );
@@ -207,7 +240,7 @@ char *par_mktmpdir ( char **argv ) {
             sprintf(
                 stmpdir,
                 "%s%scache-%s%s",
-                stmpdir2, dir_sep, buf, subdirbuf_suffix
+                top_tmpdir, dir_sep, buf, subdirbuf_suffix
             );
         }
         else {
@@ -228,7 +261,7 @@ char *par_mktmpdir ( char **argv ) {
             sprintf(
                 stmpdir,
                 "%s%scache-%s%s",
-                stmpdir2, dir_sep, sha1, subdirbuf_suffix
+                top_tmpdir, dir_sep, sha1, subdirbuf_suffix
             );
         }
     }
@@ -241,7 +274,7 @@ char *par_mktmpdir ( char **argv ) {
         sprintf(
             stmpdir,
             "%s%stemp-%u%s",
-            stmpdir2, dir_sep, getpid(), subdirbuf_suffix
+            top_tmpdir, dir_sep, getpid(), subdirbuf_suffix
         );
 
         /* Ensure we pick an unused directory each time.  If the directory
@@ -250,16 +283,16 @@ char *par_mktmpdir ( char **argv ) {
            a prior invocation crashed leaving garbage in a temp directory that
            might interfere. */
 
-        while (my_mkdir(stmpdir, 0755) == -1 && errno == EEXIST) {
+        while (my_mkdir(stmpdir, 0700) == -1 && errno == EEXIST) {
             sprintf(
                 stmpdir,
                 "%s%stemp-%u-%u%s",
-                stmpdir2, dir_sep, getpid(), ++i, subdirbuf_suffix
+                top_tmpdir, dir_sep, getpid(), ++i, subdirbuf_suffix
                 );
         }
     }
 
-    free(stmpdir2);
+    free(top_tmpdir);
 
     /* set dynamic loading path */
     par_setenv(PAR_TEMP, stmpdir);
